#!/usr/bin/env python
# $Id: pfbackup.py 172137ab6744 2012/02/23 22:55:35 patrick $
"""Backup site into a XML file."""

import sys
import logging
from os import getcwd
from os.path import exists, basename
from optparse import OptionParser
from locale import getdefaultlocale
from sqlalchemy import engine_from_config

from pyramid.paster import get_appsettings
from pyramid.threadlocal import get_current_registry

from ..lib.utils import _, localizer
from ..lib.xml import export_configuration
from ..lib.build.agent import AgentBuildManager
from ..lib.build.front import FrontBuildManager
from ..models import DBSession, Base
from ..models.users import User
from ..models.groups import Group
from ..models.storages import Storage
from ..models.projects import Project


# =============================================================================
def main():
    """Main function."""
    # Parse arguments
    parser = OptionParser('Usage: %prog <config_uri> <backup_file>\n'
        'Examples:\n  %prog production.ini pfinstance.pf.xml'
        '\n  %prog production.ini#Foo pfinstance.pf.xml --nousers')
    parser.add_option('--nousers', dest='nousers',
        help='no backup for users', default=False, action='store_true')
    parser.add_option('--nogroups', dest='nogroups',
        help='no backup for groups', default=False, action='store_true')
    parser.add_option('--nostorages', dest='nostorages',
        help='no backup for storages', default=False, action='store_true')
    parser.add_option('--noprojects', dest='noprojects',
        help='no backup for projects', default=False, action='store_true')
    parser.add_option('--loglevel', dest='log_level', help='log level',
        default='INFO')
    opts, args = parser.parse_args()
    log_level = getattr(logging, opts.log_level.upper(), None)
    if len(args) < 2 or not exists(args[0].partition('#')[0]) or not log_level:
        parser.error('Incorrect argument')
        sys.exit(2)

    logformat = '%(levelname)-8s %(message)s'
    logging.basicConfig(level=log_level, format=logformat)
    Backup(args[0]).save(opts, args[1])


# =============================================================================
class Backup(object):
    """Class to backup site."""

    # -------------------------------------------------------------------------
    def __init__(self, conf_uri):
        """Constructor method."""
        self._settings = get_appsettings(conf_uri)
        if 3 > len(self._settings):
            try:
                self._settings = get_appsettings(conf_uri, 'PubliForge')
            except LookupError:
                self._settings = get_appsettings(conf_uri, basename(getcwd()))

    # -------------------------------------------------------------------------
    def save(self, opts, filename):
        """Save asked elements.

        :param opts: (dictionary)
            List of command line options.
        :param filename: (string)
            Name of backup file.
        """
        # Initialize SqlAlchemy session
        engine = engine_from_config(self._settings, 'sqlalchemy.')
        DBSession.configure(bind=engine)
        Base.metadata.bind = engine

        # Browse elements
        elements = []
        if not opts.nousers:
            for user in DBSession.query(User):
                elements.append(user.xml(True))
        if not opts.nogroups:
            for group in DBSession.query(Group):
                elements.append(group.xml())
        if not opts.nostorages:
            for storage in DBSession.query(Storage):
                elements.append(storage.xml())
        if not opts.noprojects:
            request = DummyRequest(self._settings)
            for project in DBSession.query(Project):
                elements.append(project.xml(request))

        # Export configuration
        if elements:
            export_configuration(elements, filename, True)
        else:
            logging.warning(localizer(
                getdefaultlocale()[0]).translate(_('nothing to do!')))
        DBSession.close()


# =============================================================================
class DummyRequest(object):
    """Dummy request for processors."""
    # pylint: disable = I0011, R0903

    # -------------------------------------------------------------------------
    def __init__(self, settings):
        """Constructor method."""
        self.registry = get_current_registry()
        self.registry['abuild'] = AgentBuildManager(settings)
        self.registry['fbuild'] = FrontBuildManager(settings)
        self.registry.settings = settings
        self.session = {'lang': getdefaultlocale()[0], 'user_id': 1}



# $Id: test_views_project.py bc236c725df4 2012/01/24 22:50:00 patrick $
# pylint: disable = I0011, R0904
"""Tests of ``views.project`` functions."""

from ..tests import ModelTestCase, FunctionalTestCase


# =============================================================================
class UnitTestViewsProjectProjectView(ModelTestCase):
    """Unit test class for ``views.project.Project``."""

    # -------------------------------------------------------------------------
    def test_admin(self):
        """[u:views.project.admin]"""
        from ..views.project import ProjectView
        from ..lib.renderer import Paging
        from ..lib.form import Form
        self.config.add_route('site_admin', '/site/admin')
        result = ProjectView(self._make_request()).admin()
        self.assertTrue('form' in result)
        self.assertTrue(isinstance(result['form'], Form))
        self.assertTrue('paging' in result)
        self.assertTrue(isinstance(result['paging'], Paging))
        self.assertTrue('project_status' in result)
        self.assertTrue('active' in result['project_status'])


# =============================================================================
class FunctionalTestViewProjectView(FunctionalTestCase):
    """Functional test class for ``views.project.Project``."""

    # -------------------------------------------------------------------------
    def test_not_authenticated(self):
        """[f:views.project.admin] not authenticated."""
        response = self.testapp.get('/project/admin', status=302)
        self.assertEqual(response.location,
                         'http://localhost/login?came_from=%2Fproject%2Fadmin')

    # -------------------------------------------------------------------------
    def test_not_authorized(self):
        """[f:views.project.admin] not authorized."""
        self.testapp.login('user3')
        self.testapp.get('/project/admin', status=403)

    # -------------------------------------------------------------------------
    def test_authorized(self):
        """[f:views.project.admin] authorized."""
        self.testapp.login('user1')
        self.testapp.get('/project/admin', status=200)



# $Id: group.py 5fd521ecd19f 2012/03/19 11:02:44 patrick $
# pylint: disable = I0011, C0322
"""Group view callables."""

from colander import Mapping, SchemaNode, String
from colander import Length, OneOf

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPForbidden, HTTPFound, HTTPNotFound

from ..lib.utils import _, has_permission
from ..lib.xml import upload_configuration, export_configuration
from ..lib.views import get_action, paging_users, paging_groups
from ..lib.form import Form
from ..lib.renderer import TabSet
from ..models import ID_LEN, LABEL_LEN, DESCRIPTION_LEN
from ..models import DBSession, close_dbsession
from ..models.users import PERM_SCOPES, USER_PERMS, User
from ..models.groups import GROUPS_USERS, Group, GroupPerm
from ..models.storages import StorageUser
from ..models.projects import ProjectUser


GROUP_SETTINGS_TABS = (_('Description'), _('Permissions'), _('Members'))


# =============================================================================
class GroupView(object):
    """Class to manage groups."""

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        request.add_finished_callback(close_dbsession)
        self._request = request

    # -------------------------------------------------------------------------
    @view_config(route_name='group_admin',
                 renderer='../Templates/grp_admin.pt',
                 permission='grp.update')
    def admin(self):
        """List groups for administration purpose."""
        action, items = get_action(self._request)
        if action == 'imp!':
            upload_configuration(self._request, 'grp_manager', 'group')
            action = ''
        elif action[0:4] == 'del!':
            self._delete_groups(items)
            action = ''
        elif action[0:4] == 'exp!':
            return self._export_groups(items)

        paging, defaults = paging_groups(self._request, 'groups!')
        form = Form(self._request, defaults=defaults)

        depth = (self._request.breadcrumbs.current_path() ==
                 self._request.route_path('site_admin') and 3) or 2
        self._request.breadcrumbs.add(_('Group administration'), depth)
        return {
            'form': form, 'paging': paging, 'action': action,
            'i_manager': has_permission(self._request, 'grp_manager')}

    # -------------------------------------------------------------------------
    @view_config(route_name='group_view', renderer='../Templates/grp_view.pt',
                 permission='grp.read')
    def view(self):
        """Show group settings."""
        group_id = self._request.matchdict.get('group_id')
        action = get_action(self._request)[0]
        if action == 'exp!':
            return self._export_groups((group_id,))

        group = DBSession.query(Group).filter_by(group_id=group_id).first()
        if group is None:
            raise HTTPNotFound()
        tab_set = TabSet(self._request, GROUP_SETTINGS_TABS)
        members = [k.user_id for k in group.users]
        if len(members):
            members = DBSession.query(User.user_id, User.login, User.name)\
                .filter(User.user_id.in_(members)).order_by(User.login).all()

        self._request.breadcrumbs.add(
            _('Group settings'), replace=self._request.route_path(
            'group_edit', group_id=group.group_id))
        return {
            'form': Form(self._request), 'tab_set': tab_set,
            'PERM_SCOPES': PERM_SCOPES, 'USER_PERMS': USER_PERMS,
            'group': group, 'members': members,
            'i_editor': has_permission(self._request, 'grp_editor')}

    # -------------------------------------------------------------------------
    @view_config(route_name='group_create',
                 renderer='../Templates/grp_edit.pt',
                 permission='grp.create')
    def create(self):
        """Create a group."""
        form, tab_set = self._settings_form()

        if form.validate():
            group = self._create(form.values)
            if group is not None:
                self._request.breadcrumbs.pop()
                return HTTPFound(self._request.route_path(
                    'group_edit', group_id=group.group_id))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        self._request.breadcrumbs.add(_('Group creation'))
        return {
            'form': form, 'tab_set': tab_set,
            'PERM_SCOPES': PERM_SCOPES, 'USER_PERMS': USER_PERMS,
             'group': None, 'members': None, 'paging': None, 'groups': None}

    # -------------------------------------------------------------------------
    @view_config(route_name='group_edit', renderer='../Templates/grp_edit.pt',
                 permission='grp.update')
    def edit(self):
        """Edit group settings."""
        # Action
        paging = groups = None
        group_id = self._request.matchdict.get('group_id')
        action = get_action(self._request)[0]
        if action == 'aur?' \
               or (not action and 'paging_id' in self._request.GET):
            paging = paging_users(self._request)[0]
            groups = DBSession.query(Group.group_id, Group.label).all()
        elif action[0:4] == 'aur!':
            self._add_members(group_id)
        elif action[0:4] == 'rur!':
            DBSession.execute(GROUPS_USERS.delete()
                .where(GROUPS_USERS.c.group_id == group_id)
                .where(GROUPS_USERS.c.user_id == int(action[4:])))
            DBSession.query(StorageUser).filter_by(user_id=int(action[4:]))\
                .filter(StorageUser.perm != None).delete()
            DBSession.query(ProjectUser).filter_by(user_id=int(action[4:]))\
                .filter(ProjectUser.perm != None).delete()
            DBSession.commit()
            del self._request.session['menu']

        # Environment
        group = DBSession.query(Group).filter_by(group_id=group_id).first()
        if group is None:
            raise HTTPNotFound()
        form, tab_set = self._settings_form(group)
        members = [k.user_id for k in group.users]
        if len(members):
            members = DBSession.query(User.user_id, User.login, User.name)\
                .filter(User.user_id.in_(members)).order_by(User.login).all()

        # Save
        view_path = self._request.route_path('group_view',
            group_id=group.group_id)
        if action == 'sav!' and form.validate(group) \
               and self._save(group, form.values):
            return HTTPFound(view_path)
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Breadcrumbs
        self._request.breadcrumbs.add(_('Group settings'), replace=view_path)

        return {
            'form': form, 'action': action, 'tab_set': tab_set,
            'PERM_SCOPES': PERM_SCOPES, 'USER_PERMS': USER_PERMS,
            'group': group, 'members': members, 'paging': paging,
            'groups': groups}

    # -------------------------------------------------------------------------
    def _delete_groups(self, group_ids):
        """Delete groups.

        :param group_ids: (list)
            List of group IDs to delete.
        """
        if not has_permission(self._request, 'grp_manager'):
            raise HTTPForbidden()
        group_ids = set(group_ids)

        # Do not delete my own groups
        user_id = self._request.session['user_id']
        ids = set([k[0] for k in DBSession.query(GROUPS_USERS.c.group_id)
                   .filter(User.user_id == user_id)
                   .filter(User.user_id == GROUPS_USERS.c.user_id).all()])
        if group_ids & ids:
            self._request.session.flash(
                _("You can't delete your own group."), 'alert')
            return

        # Delete
        DBSession.query(Group).filter(
            Group.group_id.in_(group_ids)).delete('fetch')
        DBSession.commit()

    # -------------------------------------------------------------------------
    def _export_groups(self, group_ids):
        """Export groups.

        :param group_ids: (list)
            List of group IDs to export.
        :return: (:class:`pyramid.response.Response` instance)
        """
        if not has_permission(self._request, 'grp_user'):
            raise HTTPForbidden()
        elements = []
        for group in DBSession.query(Group)\
                .filter(Group.group_id.in_(group_ids)).order_by('label'):
            elements.append(group.xml())

        name = '%s_groups.pfg' % self._request.registry.settings.get(
            'skin.label', 'publiforge')
        return export_configuration(elements, name)

    # -------------------------------------------------------------------------
    def _settings_form(self, group=None):
        """Return a group settings form.

        :param group: (:class:`~..models.groups.Group` instance, optional)
            Current group object.
        :return: (tuple)
            A tuple such as ``(form, tab_set)``.
        """
        # Schema
        schema = SchemaNode(Mapping())
        if group is None:
            schema.add(SchemaNode(String(), name='group_id',
                validator=Length(max=ID_LEN)))
        schema.add(SchemaNode(String(), name='description',
            validator=Length(max=DESCRIPTION_LEN), missing=''))
        schema.add(SchemaNode(String(), name='label',
            validator=Length(min=2, max=LABEL_LEN)))
        for scope in PERM_SCOPES:
            schema.add(SchemaNode(String(), name='perm_%s' % scope,
                validator=OneOf(USER_PERMS.keys()), missing='-'))

        # Defaults
        defaults = {}
        if group is not None:
            for perm in group.perms:
                defaults['perm_%s' % perm.scope] = perm.level

        return (
            Form(self._request, schema=schema, defaults=defaults, obj=group),
            TabSet(self._request, GROUP_SETTINGS_TABS))

    # -------------------------------------------------------------------------
    def _create(self, values):
        """Create a group.

        :param values: (dictionary)
            Form values.
        :return: (:class:`~..models.groups.Group` instance or ``None``)
        """
        # Check existence
        group_id = values['group_id'].strip()
        group = DBSession.query(Group).filter_by(group_id=group_id).first()
        if group is None:
            label = u' '.join(values['label'].strip().split())[0:LABEL_LEN]
            group = DBSession.query(Group).filter_by(label=label).first()
        if group is not None:
            self._request.session.flash(
                _('This group already exists.'), 'alert')
            return

        # Create group
        group = Group(group_id, label, values['description'])
        DBSession.add(group)
        DBSession.commit()

        # Add permissions
        for scope in PERM_SCOPES:
            scope = 'perm_%s' % scope
            if values[scope] != '-':
                group.perms.append(GroupPerm(
                    group.group_id, scope[5:], values[scope]))

        DBSession.commit()
        return group

    # -------------------------------------------------------------------------
    @classmethod
    def _save(cls, group, values):
        """Save a group settings.

        :param group: (:class:`~..models.groups.Group` instance)
            Group to update.
        :param values: (dictionary)
            Form values.
        :return: (boolean)
            ``True`` if succeeds.
        """
        # Update permissions
        perms = dict([(k.scope, k) for k in group.perms])
        for scope in PERM_SCOPES:
            if values['perm_%s' % scope] != '-':
                if not scope in perms:
                    group.perms.append(GroupPerm(
                        group.group_id, scope, values['perm_%s' % scope]))
                else:
                    perms[scope].level = values['perm_%s' % scope]
            elif scope in perms:
                DBSession.delete(perms[scope])

        DBSession.commit()
        return True

    # -------------------------------------------------------------------------
    def _add_members(self, group_id):
        """Add selected users to the group.

        :param group_id: (string)
            Group ID.
        """
        group = DBSession.query(Group).filter_by(
            group_id=group_id).first()
        user_ids = [k.user_id for k in group.users]
        for user_id in get_action(self._request)[1]:
            user_id = int(user_id)
            if not user_id in user_ids:
                DBSession.execute(GROUPS_USERS.insert().values(
                    group_id=group_id, user_id=user_id))
        DBSession.commit()
        del self._request.session['menu']



# $Id: containers.py 2846fc9c22b4 2012/02/23 07:19:21 patrick $
"""Container factory management."""

from os import walk, remove
from os.path import exists, join, relpath, isdir
from shutil import rmtree
import zipfile
from subprocess import Popen, PIPE

from ...utils import EXCLUDED_FILES


# =============================================================================
class ZipFactory(object):
    """Class for Zip container."""

    uid = 'Zip'

    # -------------------------------------------------------------------------
    def __init__(self, engine):
        """Constructor method.

        :param engine: (:class:`~.lib.engine.leprisme.Engine` instance)
            Engine object on which it depends.
        """
        self._engine = engine

    # -------------------------------------------------------------------------
    def make(self, fid, output):
        """Make an ePub file.

        :param fid: (string)
            File identifier.
        :param output: (string)
            Full path to output directory.
        :return: (string)
            Full path to zip file.
        """
        filename = join(output, '%s.zip' % fid)
        zip_file = zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED)
        keeptmp = self._engine.build.processing['variables'].get('keeptmp')
        for path, dirs, files in walk(output):
            for name in dirs:
                if name in EXCLUDED_FILES or '~' in name:
                    dirs.remove(name)
            for name in files:
                if not name in EXCLUDED_FILES and not '~' in name \
                       and not name.endswith('.zip'):
                    fullname = join(path, name)
                    zip_file.write(fullname, relpath(fullname, output))
                    if not keeptmp:
                        if isdir(fullname):
                            rmtree(fullname)
                        else:
                            remove(fullname)
        zip_file.close()
        return filename


# =============================================================================
class OcfFactory(object):
    """Class for Open Container Format (OCF) file."""

    uid = 'OCF'

    # -------------------------------------------------------------------------
    def __init__(self, engine):
        """Constructor method.

        :param engine: (:class:`~.lib.engine.leprisme.Engine` instance)
            Engine object on which it depends.
        """
        self._engine = engine

    # -------------------------------------------------------------------------
    def make(self, fid, output):
        """Make an Open Container Format file and check its validity.

        :param fid: (string)
            File identifier.
        :param output: (string)
            Full path to output directory.
        :return: (string)
            Full path to ePub file.
        """
        # Input directory
        root = self._engine.config('container:%s' % self.uid, 'root', '')
        if not exists(join(output, root)):
            self._engine.build.stopped('Unknown file "%s"' % root, 'a_error')
            return
        if not exists(join(output, root, 'mimetype')) \
               or not exists(join(output, root, 'META-INF', 'container.xml')):
            self._engine.build.stopped(
                '%s: incorrect OCF structure' % root, 'a_error')
            return
        root = join(output, root)

        # Create zip file
        ocf_file = join(output,
            self._engine.config('Output', 'format').format(fid=fid))
        zip_file = zipfile.ZipFile(ocf_file, 'w', zipfile.ZIP_DEFLATED)
        zip_file.write(join(root, 'mimetype'), 'mimetype', zipfile.ZIP_STORED)
        for path, dirs, files in walk(root):
            for name in dirs:
                if name in EXCLUDED_FILES or '~' in name:
                    dirs.remove(name)
            for name in files:
                if not name in EXCLUDED_FILES  and not '~' in name \
                       and name != 'mimetype':
                    zip_file.write(
                        join(path, name), relpath(join(path, name), root))
        zip_file.close()
        if self._engine.build.stopped():
            return

        # OCF checker
        name = self._engine.config('container:%s' % self.uid, 'checker')
        if not name or \
               not self._engine.build.processing['variables'].get('ocfcheck'):
            return ocf_file

        err = Popen([k.format(ocffile=ocf_file) for k in name.split()],
                   stdout=PIPE, stderr=PIPE).communicate()[1]
        if err:
            remove(ocf_file)
            err = err.replace(output.encode('utf8'), '') .replace('\n', ' ')
            self._engine.build.stopped(err.decode('utf8'), 'a_error')
            return

        return ocf_file



# $Id: front.py ada86c3ea608 2012/03/22 22:52:47 patrick $
"""Front build management.

A *front* is a web site.
"""
# pylint: disable = I0011, C0302

import logging
import xmlrpclib
from time import time, sleep
from random import randint
from threading import Thread
from os import walk, makedirs, rename
from os.path import join, exists, isfile, dirname, relpath, normpath
from cStringIO import StringIO
from lxml import etree
import re

from pyramid.i18n import get_localizer

from ..utils import _, localizer, copy_content, decrypt
from ..utils import camel_case, has_permission, EXCLUDED_FILES
from ..rsync import get_block_size
from ..rsync import SigFile, PatchedFile, DeltaFile
from ...views import xmlrpc
from ...models import DBSession
from ...models.users import User
from ...models.processors import Processor
from ...models.storages import Storage, StorageUser


LOG = logging.getLogger(__name__)


# =============================================================================
class FrontBuildManager(object):
    """This class manages front builds.

    One instance of :class:`FrontBuildManager` is created during application
    initialization. It is only used in front mode. It is stored in application
    registry.

    ``self._agents`` is a dictionary such as ``{url: (password, weight,
    processor_list, processor_expire_time),... }`` which stores agent features.

    ``self._builds`` is a dictionary of :class:`FrontBuild` objects.

    ``self._results`` is a dictionary of dictionaries such as ``{build_id:
    result_dict}``. ``result_dict`` is a dictionary with following keys:
    ``status``, ``log``, ``expire``, ``project_id``, ``user_id``. According to
    build events, it can also contains ``files``, ``values`` and ``error``
    keys.

    ``self._results[build_id]['status']`` is one of the following strings:
    ``stop``, ``fatal`` or ``end``.

    ``self._results[build_id]['log']`` is a list of tuples such as
    ``(timestamp, step, percent, message)``.
    """
    # pylint: disable = I0011, R0902

    # -------------------------------------------------------------------------
    def __init__(self, settings):
        """Constructor method.

        :param settings: (dictionary)
            Setting dictionary.
        """
        # Attributes
        self.storage_root = settings['storage.root']
        self.build_ttl = int(settings.get('build.ttl', 1800))
        self.result_ttl = int(settings.get('build.result_ttl', 86400))
        self._concurrent = int(settings.get('build.concurrent', 3))
        self._refresh = int(settings.get('agent.refresh', 0))
        self._builds = {}
        self._results = {}

        # Agent list
        self._agents = {}
        total = 0
        for index in range(0, 100):
            pfx = 'agent.%d' % index
            weight = int(settings.get('%s.weight' % pfx, 0))
            if weight:
                self._agents[settings.get('%s.url' % pfx, '')] = [
                    settings.get('%s.password' % pfx, ''), weight, None, 0]
                total += weight
        if total:
            self._concurrent = int(self._concurrent * total
                / float(min([self._agents[k][1] for k in self._agents])))

    # -------------------------------------------------------------------------
    def agent_list(self):
        """Return a list of available agents."""
        return self._agents.keys()

    # -------------------------------------------------------------------------
    def refresh_agent_list(self, request):
        """Refresh processor list for each agent.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        """
        # Refresh processor list
        now = time()
        for url in self._agents:
            if self._agents[url][2] is None \
                   or (self._refresh and self._agents[url][3] < now):
                error, processors = self.call(request, url, 'processor_list')
                if error:
                    LOG.error('%s: %s', url, error)
                self._agents[url][2] = processors or []
                self._agents[url][3] = now + self._refresh

        # Refresh processor records
        if self._refresh:
            DBSession.query(Processor).filter(
                Processor.updated + self._refresh < now).delete()
            DBSession.commit()
        processors = [k[0] for k in DBSession.query(Processor.processor_id)]
        for url in sorted(self._agents):
            for processor_id in self._agents[url][2]:
                if processor_id in processors:
                    continue
                error, xml = self.call(
                    request, url, 'processor_xml', processor_id)
                if error:
                    LOG.error('%s: %s', url, error)
                    continue
                processor = Processor.load(processor_id, xml)
                if isinstance(processor, basestring):
                    LOG.error('%s: %s', url,
                              get_localizer(request).translate(processor))
                    continue
                DBSession.add(processor)
                DBSession.commit()

    # -------------------------------------------------------------------------
    def processor(self, request, processor_id):
        """Return a processor tree.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param processor_id: (string)
            Processor ID
        :return: (:class:lxml.etree.ElementTree or ``None``)
        """
        self.refresh_agent_list(request)
        xml = DBSession.query(Processor.xml).filter_by(
            processor_id=processor_id).first()
        if xml is not None:
            return etree.parse(StringIO(xml[0].encode('utf8')))

    # -------------------------------------------------------------------------
    def start_build(self, request, processing, processor, pack):
        """Find an agent, convert processing and pack into dictionaries,
        create a build, add it in ``self._builds`` dictionary and try to start
        it.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param processing:
            (:class:`~.models.projects.ProjectProcessing` instance)
            Processing object.
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :param pack: (:class:`~.models.projects.ProjectPack` instance)
            Pack object.
        :return: (:class:`FrontBuild` instance)
        """
        # Compute ID
        self._cleanup()
        build_id = '{prj_id}-{prc_id}-{pck_id}'.format(
            prj_id=processing.project_id, prc_id=processing.processing_id,
            pck_id=pack.pack_id)
        if build_id in self._builds:
            request.session.flash(_('${i}: action already in progress.',
                {'i': build_id}), 'alert')
            return

        # Find agent URL
        agent_url = self._find_url(request, processing.processor)
        if agent_url is None:
            request.session.flash(_('No available agent for ${p}.',
                {'p': processing.processor}), 'alert')
            return

        # Processing dictionary
        processing_dict = {'processor_id': processing.processor,
            'variables': self._variables2dict(request, processing, processor),
            'resources': self._file_set2list(request, processing, 'resource'),
            'templates': self._file_set2list(request, processing, 'template'),
            'output': processing.output}
        if request.session.peek_flash('alert'):
            return

        # Pack dictionary
        pack_dict = {'recursive': pack.recursive,
            'files': self._file_set2list(request, pack, 'file'),
            'resources': self._file_set2list(request, pack, 'resource'),
            'templates': self._file_set2list(request, pack, 'template')}
        if request.session.peek_flash('alert'):
            return

        # Context
        context = {'lang': request.session['lang'],
                   'project_id': request.session['project']['project_id'],
                   'user_id': request.session['user_id']}

        # Create an FrontBuild object
        if build_id in self._results:
            del self._results[build_id]
        self._builds[build_id] = FrontBuild(
            self, build_id, agent_url, context, processing_dict, pack_dict)
        self.start_waiting(request)
        return self._builds[build_id]

    # -------------------------------------------------------------------------
    def build_list(self, project_id):
        """List all builds of a project.

        :param project_id: (integer)
            Project ID.
        :return: (list)
            A list of dictionaries.

        Returned dictionaries are sorted by start time. Each dictionary has
        following keys: ``build_id``, ``start``, ``status``, ``processing_id``,
        ``pack_id``, ``user_id``.
        """
        self._cleanup()
        builds = []
        for build_id, build in self._builds.items():
            if build.result['project_id'] == project_id:
                builds.append({
                    'build_id': build_id, 'start': build.result['log'][0][0],
                    'status': build.result['status'],
                    'processing_id': int(build_id.split('-')[1]),
                    'pack_id': int(build_id.split('-')[2]),
                    'user_id': build.result['user_id']})

        for build_id, result in self._results.items():
            if result['project_id'] == project_id:
                builds.append({
                    'build_id': build_id, 'start': result['log'][0][0],
                    'status': result['status'],
                    'processing_id': int(build_id.split('-')[1]),
                    'pack_id': int(build_id.split('-')[2]),
                    'user_id': result['user_id'],
                    'files': 'files' in result and result['files'] or None})

        return sorted(builds, key=lambda k: k['start'], reverse=True)

    # -------------------------------------------------------------------------
    def is_owner(self, request, build_id, user_id=None):
        """Check if user ``user_id`` has launched the build ``build_id``.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param build_id: (string)
            Build to check
        :param user_id: (integer, optional)
            User ID to check. By default, the current user is checked.
        :rtype: (boolean)
        """
        if user_id is None:
            user_id = request.session['user_id']
        if (build_id in self._results
                and self._results[build_id]['user_id'] == user_id) \
                or (build_id in self._builds and
                    self._builds[build_id].result['user_id'] == user_id)\
                or has_permission(request, 'prj_manager'):
            return True
        return False

    # -------------------------------------------------------------------------
    def progress(self, request, build_ids):
        """Return the progress of builds.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param build_ids: (list)
            Build ID list.
        :return: (tuple)
            Return a tuple such as ``(<working>, <progress_dictionary>)``.

        ``<working>`` is a boolean indicating whether one of the processing is
         in progress.

         ``<progress_dictionary>`` is a dictionary such as ``{<build_id>:
         (<step>, <percent>, <message>),...}`` where ``<step>`` is one of the
         following:

        * ``wait``: waiting
        * ``start``: starting
        * ``sync``: synchronizing storages between front and agent
        * ``a_???``: an :class:`~.lib.build.agent.AgentBuildManager` step
        * ``get``: getting result
        * ``warn``: a warning occurred
        * ``error``: an error occurred
        * ``fatal``: a fatal error occurred
        * ``stop``: stopping
        * ``end``: successfully completed
        * ``none``: unknown or not in progress build
        """
        self._cleanup()
        working = False
        prgrss = {}
        for build_id in build_ids:
            if build_id in self._builds:
                prgrss[build_id] = self._builds[build_id].progress(request)
                working = True
            else:
                prgrss[build_id] = ('none', 0, '')
        self.start_waiting(request)
        return working, prgrss

    # -------------------------------------------------------------------------
    def stop(self, request, build_id):
        """Stop a build.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param build_id: (string)
            Build ID.
        """
        self._cleanup()
        if build_id in self._builds:
            self._builds[build_id].stop(request)

    # -------------------------------------------------------------------------
    def complete(self, request, build_id, key):
        """Get the result and eventually download the output directory in
        storage.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param build_id: (string)
            Build ID.
        :param key: (string)
            Key to authenticate the request.
        :return: (boolean)
        """
        self._cleanup()
        if build_id in self._builds:
            return self._builds[build_id].complete(request, key)
        return build_id in self._results

    # -------------------------------------------------------------------------
    def result(self, request, build_id):
        """Return the result of a build.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param build_id: (string)
            Build ID.
        :return: (tuple)
            A tuple such as ``(<status>, <values>, <files>, <log>)``.

        The status ``<status>`` is one of the following:

        * ``stop``: stopped
        * ``fatal``: in error
        * ``end``: successfuly completed
        """
        self._cleanup()
        if not 'project' in request.session:
            return {}
        project_id = request.session['project']['project_id']
        if build_id in self._results \
               and self._results[build_id]['project_id'] == project_id:
            return self._results[build_id]
        return {}

   # -------------------------------------------------------------------------
    def call(self, request, url, method, *args):
        """Call an agent method directly or via RPC.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param url: (string)
            The agent URL or ``localhost`` to call without RPC.
        :param method: (string)
            Method to call
        :param args:
            Non-keyworded arguments for method.
        :return: (tuple)
            A tuple such as ``(<error>, <result>)`` where ``<error>`` is a
            string and ``<result>`` depends on ``method``.

        In addition to the required arguments, this method sends also a context
        dictionary with ``lang`` (language for error messages), ``front_id``,
        ``password`` (to authenticate front), ``user_id`` and ``local``
        (``True`` if called without XML-RPC). If this method is called for a
        local agent, it adds a ``request`` key in context.
        """
        # Create context
        context = {'lang': request.session['lang'],
                   'front_id': request.registry.settings['uid'],
                   'password': self._agents[url][0],
                   'user_id': request.session['user_id']}

        # Local agent
        if not url:
            context['local'] = True
            context['request'] = request
            try:
                return getattr(xmlrpc, method)(request, context, *args)
            except AttributeError, err:
                return err, None

        # Remote agent
        context['local'] = False
        proxy = xmlrpclib.ServerProxy('%s/xmlrpc' % url, verbose=False)
        try:
            error, result = getattr(proxy, method)(context, *args)
        except IOError, err:
            error, result = err.strerror, None
        except OverflowError, err:
            error, result = err, None
        except xmlrpclib.ProtocolError, err:
            error, result = err.errmsg, None
        except xmlrpclib.Fault, err:
            error, result = err.faultString, None
        return error, result

    # -------------------------------------------------------------------------
    def start_waiting(self, request):
        """Start waiting builds.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        """
        # Check waiting builds
        waiting = []
        running = 0
        for build_id in self._builds:
            if self._builds[build_id].result['log'][0][1] == 'wait':
                waiting.append(build_id)
            elif not self._builds[build_id].stopped():
                running += 1
        if not waiting or running >= self._concurrent:
            return

        # Start waiting
        waiting = sorted(waiting,
            key=lambda build_id: self._builds[build_id].result['log'][0][0])
        for build_id in waiting[0:self._concurrent - running]:
            self._builds[build_id].start(request)
            sleep(1)
        for build_id in waiting[self._concurrent - running:]:
            self._builds[build_id].expire = time() + self.build_ttl

    # -------------------------------------------------------------------------
    def _cleanup(self):
        """Delete completed builds and expired results and kill long builds."""
        # Build -> result or stop
        now = time()
        for build_id in self._builds.keys():
            if self._builds[build_id].stopped():
                self._results[build_id] = self._builds[build_id].result
                self._results[build_id]['expire'] = now + self.result_ttl
                del self._builds[build_id]
            elif self._builds[build_id].expire < now:
                self._builds[build_id].stop()

        # Remove old results
        for build_id in self._results.keys():
            if self._results[build_id]['expire'] < now:
                del self._results[build_id]

    # -------------------------------------------------------------------------
    def _find_url(self, request, processor_id):
        """Find URL of an agent which serves ``processor_id``.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param processor_id: (string)
            ID of the processor agent must serve.
        :return: (string):
            URL of the found agent or ``None``.
        """
        found_url = None
        min_activity = 0
        self.refresh_agent_list(request)
        for url in self._agents:
            if self._agents[url][2] and processor_id in self._agents[url][2]:
                error, activity = self.call(request, url, 'activity')
                if error:
                    LOG.error('%s: %s', url, error)
                    continue
                activity = float(activity + 1) / self._agents[url][1]
                if found_url is None or activity < min_activity:
                    found_url, min_activity = url, activity
        return found_url

    # -------------------------------------------------------------------------
    @classmethod
    def _variables2dict(cls, request, processing, processor):
        """Create a variable dictionary from a processing record and its
        processor.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param processing:
            (:class:`~.models.projects.ProjectProcessing` instance)
        :param processor: (:class:`lxml.etree.ElementTree` instance)
            Processor of current processing.
        :return: (dictionary)
        """
        values = dict([(k.name, k) for k in processing.variables])
        variables = {}
        for var in processor.findall('processor/variables/group/var'):
            name = var.get('name')
            value = values[name].value if name in values \
                    else var.findtext('default') is not None \
                    and var.findtext('default').strip() or ''
            if var.get('type') == 'string':
                variables[name] = value
            elif var.get('type') == 'boolean':
                variables[name] = bool(value == 'true' or value == '1')
            elif var.get('type') == 'integer':
                variables[name] = int(value)
            elif var.get('type') == 'select':
                if not value in [k.get('value') or k.text
                                 for k in var.findall('option')]:
                    request.session.flash(
                        _('${v}: bad value.', {'v': name}), 'alert')
                    return variables
                variables[name] = int(value) if value.isdigit() else value
            elif var.get('type') == 'regex':
                if not re.match(var.find('pattern').text, value):
                    request.session.flash(
                        _('${v}: bad value.', {'v': name}), 'alert')
                    return variables
                variables[name] = int(value) if value.isdigit() else value
        return variables

    # -------------------------------------------------------------------------
    @classmethod
    def _file_set2list(cls, request, record, file_type):
        """Save set of files in a list.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param record: (:class:`~.models.projects.ProjectProcessing`
            or :class:`~.models.projects.ProjectPack` instance).
        :param file_type: ('file', 'resource' or 'template')
            File type.
        """
        file_set = []
        items = [k for k in record.files if k.file_type == file_type]
        if len(items) == 0:
            return file_set

        storage_root = request.registry.settings['storage.root']
        for item in items:
            if not exists(join(storage_root, item.path)):
                request.session.flash(
                    _('"${n}" does not exists.', {'n': item.path}), 'alert')
                return
            if file_type in ('file', 'resource'):
                file_set.append(item.path)
            else:
                file_set.append((item.path, item.target))

        return file_set


# =============================================================================
class FrontBuild(object):
    """This class manages a build locally and via an agent.

    ``self.result`` is a dictionary with following keys: ``status``, ``log``,
    ``start``, ``expire``, ``project_id``, ``user_id``. At the process end, it
    can also have ``files``, ``values``, ``error`` and ``end`` keys.

    ``self.result['log']`` is a list of tuples such as ``(timestamp, step,
    percent, message)``.

    ``self.result['expire']`` is the date beyond which the build is destroyed.

    ``self.key`` is a key to authenticate transaction between front and
    agent.
    """
    # pylint: disable = I0011, R0902, R0913

    # -------------------------------------------------------------------------
    def __init__(self, build_manager, build_id, agent_url, context,
                 processing, pack):
        """Constructor method.

        :param build_manager: (:class:`FrontBuildManager` instance)
            Application :class:`FrontBuildManager` object.
        :param build_id: (string)
            Build ID.
        :param agent_url: (string)
            URL to call to complete process.
        :param context: (dictionary)
            A dictionary with with ``project_id``, ``lang`` and ``user_id``
            keys.
        :param processing: (dictionary)
            A processing dictionary.
        :param pack: (dictionary)
            A pack dictionary.
         """
        self.uid = build_id
        self._build_manager = build_manager
        self._agent_url = agent_url
        self._processing = processing
        self._pack = pack
        self._lang = context['lang']
        self._thread = None
        self.key = str(randint(1000, 9999999))
        self.expire = time() + build_manager.build_ttl
        self.result = {'status': 'none',
            'log': [(time(), 'wait', 1, self._translate(_('waiting...')))],
            'project_id': context['project_id'], 'user_id': context['user_id']}

    # -------------------------------------------------------------------------
    def start(self, request):
        """Start a build in a thread.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        """
        if self._thread is None:
            self._thread = Thread(
                target=self._thread_start, kwargs={'request': request})
            self.result['log'] = [
                (time(), 'start', 1, self._translate(_('startup')))]
            self._thread.start()

    # -------------------------------------------------------------------------
    def progress(self, request):
        """Return the progress of build.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :return: (tuple)
            A tuple such as ``(<step>, <percent>, <message>, <start_time>)``.
        """
        if self.result['log'][-1][1][0:2] != 'a_':
            return self.result['log'][-1][1:] + (self.result['log'][0][0],)

        error, prgrss = self._build_manager.call(
            request, self._agent_url, 'progress', self.uid)
        if error:
            self.stopped(error, 'a_error')
            return self.result['log'][-1][1:] + (self.result['log'][0][0],)
        if prgrss[0] == 'none':
            self.stopped(_('agent build destroyed'))
            prgrss = self.result['log'][-1][1:]
        elif prgrss[0] in ('a_end', 'a_stop', 'a_fatal'):
            self.complete(request, self.key)
            prgrss = self.result['log'][-1][1:]
        return tuple(prgrss) + (self.result['log'][0][0],)

    # -------------------------------------------------------------------------
    def stop(self, request=None):
        """Stop building."""
        if request is not None and self.result['log'][-1][1][0:2] == 'a_':
            error = self._build_manager.call(
                request, self._agent_url, 'stop', self.uid)[0]
            if error:
                self.stopped(error)
        elif not self.stopped():
            self._log(_('stopped'), 'stop', 100)
            self.result['message'] = _('Stopped by user')
            self.result['status'] = 'stop'

    # -------------------------------------------------------------------------
    def stopped(self, error=None, level='fatal'):
        """Check if there is a fatal error and if the build is stopped.

        :param error: (string, optional)
            Error message.
        :param level: (string, default='fatal')
            Error level: ``warn``, ``error`` or ``fatal``.
        :return: (boolean)
            ``True`` if it is stopped.
        """
        if error:
            if level == 'fatal':
                self.result['status'] = level
                self.result['message'] = error
            self._log(error, level, 100)

        return self.result['status'] in ('stop', 'fatal', 'end')

    # -------------------------------------------------------------------------
    def complete(self, request, key):
        """Start a *complete* action in a thread.

        :param request: (object)
            WebOb request object.
        :param key: (string)
            Authentication key.
        :return: (boolean)
            ``True`` if succeeds.
        """
        if self.result['log'][-1][1][0:2] != 'a_' or self.key != key or \
               (self._thread is not None and self._thread.is_alive()):
            return False
        self._log(_('getting log and result'), 'get')

        # Update request
        user = DBSession.query(User)\
               .filter_by(user_id=int(self.result['user_id'])).first()
        if user is None:
            self.stopped(_('unknown user'))
            return False
        user.setup_environment(request)
        DBSession.close()

        # Launch thread
        if self._thread is not None:
            del self._thread
        self._thread = Thread(
            target=self._thread_complete, kwargs={'request': request})
        self._thread.start()
        return True

    # -------------------------------------------------------------------------
    def _log(self, message, step=None, percent=None):
        """Append an entry to ``result['log']``.

        :param message: (string)
            Message to write to the log.
        :param step: (string, optional)
            If not ``None``, progress is updated.
        :param percent: (int, optional)
            Percent of progress for step ``<step>``.

        ``self.result['log']`` is a list of tuples such as ``(<timestamp>,
        <step>, <percent>, <message>)``.
        """
        if percent is None:
            percent = self.result['log'][-1][2] if step is None else 0

        if step is None:
            step = self.result['log'][-1][1]

        self.result['log'].append(
            (time(), step, percent, self._translate(message)))

    # -------------------------------------------------------------------------
    def _thread_start(self, request):
        """Action launched in a thread to start a build.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        """
        # Synchronize storages with buildspaces
        if self._agent_url:
            self._storage2buildspace(request)
        if self.stopped():
            return

        # Start build
        end_url = request.route_url(
            'build_complete', build_id=self.uid, key=self.key)
        self.expire = time() + self._build_manager.build_ttl
        if self._agent_url:
            self._log(_('call ${url}', {'url': self._agent_url}), 'a_call')
        else:
            self._log(_('call agent'), 'a_call')
        error = self._build_manager.call(request, self._agent_url, 'start',
            self.uid, self._processing, self._pack, end_url)[0]
        self.stopped(error)

    # -------------------------------------------------------------------------
    def _thread_complete(self, request):
        """Action launched in a thread to get the result and, eventually,
        download the output directory in storage.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        """
        # Get result
        error, result = self._build_manager.call(
            request, self._agent_url, 'result', self.uid)
        if self.stopped(error):
            self._build_manager.start_waiting(request)
            return
        self.result['log'][-1:1] += result['log']
        if result['status'] == 'none':
            self.stopped(_('agent build destroyed'))
            self._build_manager.start_waiting(request)
            return
        elif result['status'] != 'a_end':
            self.result['status'] = result['status'][2:]
            self.result['message'] = result['message']
            self._build_manager.start_waiting(request)
            return

        # Transfer result
        if 'values' in result:
            self.result['values'] = result['values']
        if 'files' in result:
            self.result['files'] = result['files']

        # Download output directory in storage
        if self.result.get('files') and self._processing['output']:
            self._output2storage(request)
            if self.stopped():
                del self.result['files']
                self._build_manager.start_waiting(request)
                return

        # End
        if [True for k in self.result['log'] if k[1] in ('error', 'a_error')]:
            self.stopped(_('error occurred'))
            self._build_manager.start_waiting(request)
            return
        self._log(_('successfully completed'), 'end', 100)
        self.result['message'] = _('Successfully completed')
        self.result['status'] = 'end'
        self._build_manager.start_waiting(request)

    # -------------------------------------------------------------------------
    def _storage2buildspace(self, request):
        """Synchronize storages on front with buildspaces on agent.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        """
        # Make list of files to synchronize
        file_list = list(self._processing['resources']) \
                    + [k[0] for k in self._processing['templates']] \
                    + list(self._pack['files']) \
                    + list(self._pack['resources']) \
                    + [k[0] for k in self._pack['templates']]
        total = len(file_list)

        # Browse names and synchronize
        for index, name in enumerate(file_list):
            self._log(_('synchronizing ${n}', {'n': name}),
                      'sync', 100 * index / total)
            fullname = join(request.registry.settings['storage.root'], name)
            if not exists(fullname):
                continue
            if isfile(fullname):
                self._file2buildspace(request, name)
            else:
                self._dir2buildspace(request, name)
            if self.stopped():
                return

        self._log(_('synchronization completed'), 'sync', 100)
        return True

    # -------------------------------------------------------------------------
    def _dir2buildspace(self, request, directory):
        """Synchronize a directory in storage on front with its copy in
        buildspace on agent.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param directory: (string)
            Relative path to storage directory of directory to synchronize.
        """
        # Get directory content
        fullpath = join(request.registry.settings['storage.root'],
                        directory)
        file_list = []
        for path, dirs, files in walk(fullpath):
            for name in set(dirs) & set(EXCLUDED_FILES):
                dirs.remove(name)
            for name in files:
                file_list.append(relpath(join(path, name), fullpath))

        # In destination, remove deleted files
        error = self._build_manager.call(request, self._agent_url,
            'buildspace_cleanup', directory, file_list)[0]
        if self.stopped(error):
            return

        # Synchronize files
        for name in file_list:
            self._file2buildspace(request, join(directory, name))
            if self.stopped():
                return

    # -------------------------------------------------------------------------
    def _file2buildspace(self, request, filename):
        """Synchronize a file in storage on front with its copy in buildspace
        on agent.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param filename: (string)
            Relative path to storage directory of file to synchronize.
        """
        # Get file signature
        error, sig = self._build_manager.call(
            request, self._agent_url, 'buildspace_send_signature', filename)
        if self.stopped(error):
            return
        stg_path = request.registry.settings['storage.root']

        # Transfer full file
        if not sig:
            with open(join(stg_path, filename), 'rb') as hdl:
                error = self._build_manager.call(
                    request, self._agent_url, 'buildspace_receive_file',
                    filename, xmlrpclib.Binary(hdl.read()))[0]
            self.stopped(error)
            return

        # Transfer delta
        with open(join(stg_path, filename), 'rb') as hdl:
            delta_file = DeltaFile(sig.data, hdl)
            delta_buf = delta_file.read()
            delta_file.close()
        if not delta_buf or self.stopped():
            return
        error = self._build_manager.call(
            request, self._agent_url, 'buildspace_receive_delta', filename,
            xmlrpclib.Binary(delta_buf))[0]
        self.stopped(error)

    # -------------------------------------------------------------------------
    def _output2storage(self, request):
        """Copy output on agent in storage on front.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        """
        # Authorized?
        storage, root, user = self._storage4output(request)
        if storage is None:
            DBSession.close()
            return

        # Local agent
        if not self._agent_url:
            output_dir = join(request.registry.settings['build.root'],
                camel_case(self.uid), 'Output')
            if not exists(output_dir):
                self.stopped(_('Output directory has been destroyed.'))
                DBSession.close()
                return
            copy_content(output_dir, root)
        # Remote agent
        else:
            error, files = self._build_manager.call(
                request, self._agent_url, 'output_list', self.uid)
            if self.stopped(error) or not len(files):
                DBSession.close()
                return
            total = len(files)
            for index, filename in enumerate(files):
                self._log(_('getting ${f}', {'f': filename}),
                          'get', 100 * index / total)
                self._file2storage(request, root, filename)
                if self.stopped():
                    DBSession.close()
                    return

        # Add in VCS
        handler = request.registry['handler']\
              .get_handler(storage.storage_id, storage)
        if storage.vcs_engine != 'none':
            message = get_localizer(request).translate(
                _('Produced by processing'))
            handler.add((user and user.vcs_user or None,
                     user and decrypt(user.vcs_password,
                        request.registry.settings['auth.key']),
                     request.session['name']),
                    self._processing['output'].partition('/')[2], message)
            error, message = handler.progress()
            if error == 'error':
                self.stopped(message)
        handler.cache.clear()

        DBSession.close()
        if not self.stopped():
            self._log(_('result received'), 'get', 100)

    # -------------------------------------------------------------------------
    def _file2storage(self, request, root, filename):
        """Synchronize a file in build directory on agent with one in storage
        on front.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :param root: (string)
            Absolute root path for file to synchronize.
        :param filename: (string)
            Name of the file to synchronize.
        """
        fullname = normpath(join(root, filename))
        if not fullname.startswith(root):
            self.stopped(_('Access was denied to this resource.'))
            return
        if not exists(dirname(fullname)):
            makedirs(dirname(fullname))

        # Transfer full file
        if not exists(fullname):
            error, content = self._build_manager.call(request, self._agent_url,
                'output_send_file', self.uid, filename)
            if self.stopped(error):
                return
            with open(fullname, 'wb') as hdl:
                hdl.write(content.data)
            return

        # Transfer delta
        sig_file = SigFile(open(fullname, 'rb'), get_block_size(fullname))
        sig_buf = sig_file.read()
        sig_file.close()
        error, delta = self._build_manager.call(request, self._agent_url,
            'output_send_delta', self.uid, filename, xmlrpclib.Binary(sig_buf))
        if self.stopped(error):
            return

        # Patch
        patch_file = PatchedFile(open(fullname, 'rb'), StringIO(delta.data))
        temp_name = '%s~%d~' % (fullname, randint(1, 999999))
        with open(temp_name, 'wb') as hdl:
            hdl.write(patch_file.read())
        patch_file.close()
        if exists(temp_name):
            rename(temp_name, fullname)

   # -------------------------------------------------------------------------
    def _storage4output(self, request):
        """Get the corresponding storage of an output if exists and if user
        is authorized.

        :param request: (:class:`pyramid.request.Request` instance)
            Current request.
        :return: (tuple)
            A tuple such as ``(storage, root, user)`` where ``storage`` is
            a :class:`~.models.storages.Storage` object, ``root`` the
            absolute root path for files to transfer and ``user`` a
            :class:`~.models.storage.StorageUser` object.
        """
        # Storage
        storage_id = self._processing['output'].partition('/')[0]
        storage = DBSession.query(Storage).filter_by(
            storage_id=storage_id).first()
        if storage is None:
            self.stopped(_('This storage does not exist!'))
            return None, None, None

        # Root path
        storage_root = normpath(request.registry.settings['storage.root'])
        root = normpath(join(storage_root, self._processing['output']))
        if not root.startswith(storage_root):
            self.stopped(_('Access was denied to this resource.'))
            return None, None, None

        # User
        user = DBSession.query(StorageUser).filter_by(
            storage_id=storage_id,
            user_id=request.session['user_id']).first()
        if not has_permission(request, 'stg_modifier') \
                and storage.access != 'open' \
                and (not user or not user.perm or user.perm == 'user'):
            self.stopped(_('You do not have access to this storage!'))
            return None, None, None

        if storage.vcs_engine not in ('none', 'local') \
               and not (user and user.vcs_user):
            self.stopped(_('ID and password for storage are missing.'))
            return None, None, None

        return storage, root, user

    # -------------------------------------------------------------------------
    def _translate(self, text):
        """Return ``text`` translated.

        :param text: (string)
            Text to trnaslate.
        """
        return localizer(self._lang or 'en').translate(text)



#!/usr/bin/env python
# $Id: pfbuild.py b1ffa5a75b89 2012/02/19 23:23:01 patrick $
"""Console script to build a processing."""

import sys
import logging
from os.path import exists, join, abspath, dirname, expanduser, relpath
from optparse import OptionParser
from lxml import etree
from locale import getdefaultlocale
from textwrap import fill
import re

from ..lib.utils import _, localizer
from ..lib.xml import XML_NS, load, local_text
from ..lib.build.agent import AgentBuildManager


LOG = logging.getLogger(__name__)


# =============================================================================
def main():
    """Main function."""
    # Parse arguments
    parser = OptionParser(
        'Usage: %prog [options] <build_file> [<file> <file>...]')
    parser.add_option('--list-processors', dest='list_processors',
        help='list available processors', default=False, action='store_true')
    parser.add_option('--list-variables', dest='list_variables',
        help='list variable values', default=False, action='store_true')
    parser.add_option('--show-variables', dest='show_variables', default=False,
        help='show variable parameters and values', action='store_true')
    parser.add_option('--loglevel', dest='log_level', help='log level',
        default='INFO')
    parser.add_option('--logfile', dest='log_file', help='log file')
    parser.add_option('--processor_root', dest='processor_root',
        help='processor path')
    parser.add_option('--storage_root', dest='storage_root',
        help='storage path')
    parser.add_option('--build_root', dest='build_root', help='build path')
    parser.add_option('--build_id', dest='build_id', help='build ID')
    parser.add_option('--recursive', dest='recursive',
        help='recursive parsing', default=False, action='store_true')
    opts, args = parser.parse_args()
    log_level = getattr(logging, opts.log_level.upper(), None)
    if len(args) < 1 or not exists(args[0]) or not log_level:
        parser.error('Incorrect argument')
        sys.exit(2)

    # Initialize log
    logformat = '%(asctime)s %(levelname)-8s %(message)s'
    if opts.log_file:
        logging.basicConfig(filename=expanduser(opts.log_file), filemode='w',
                            level=log_level, format=logformat)
    else:
        logging.basicConfig(level=log_level, format=logformat)

    # Build
    BuildLauncher(opts).start(args[0], args[1:])


# =============================================================================
class BuildLauncher(object):
    """Class to launch a build on command-line."""

    # -------------------------------------------------------------------------
    def __init__(self, opts):
        """Constructor method."""
        self._opts = opts
        self._relaxngs = {'publiforge':
            join(dirname(__file__), '..', 'RelaxNG', 'publiforge.rng')}
        self._processor = None

    # -------------------------------------------------------------------------
    def start(self, build_file, files):
        """Start build.

        :param build_file: (string)
             Path to build XML file.
        :param files: (list)
             List of files on command-line.
        """
        # Load build file
        build_tree = load(build_file, self._relaxngs)
        if isinstance(build_tree, basestring):
            LOG.critical(self._translate(build_tree))
            return

        # Read settings
        settings = self._read_settings(build_file, build_tree)
        if settings is None:
            return

        # Create agent
        build_manager = AgentBuildManager(settings)
        build_manager.processor_list()
        if self._opts.processor_root:
            build_manager.add_processors(expanduser(self._opts.processor_root))

        # Read processing
        processing = self._read_processing(settings, build_manager, build_tree)
        if processing is None or self._list_processors(build_manager) \
               or self._display_variables(processing):
            return

        # Read pack
        pack = self._read_pack(settings, build_tree, files)
        if pack is None:
            return

        # Create a build and start it
        build_id = self._opts.build_id  \
                   or build_tree.find('build').get('%sid' % XML_NS)
        context = {'lang': getdefaultlocale()[0], 'local': True}
        build = build_manager.start_build(build_id, context, processing, pack)

        # Display result
        if build.result['status'] != 'a_end':
            return
        if 'a_error' in [k[1] for k in build.result['log']]:
            LOG.critical(self._translate(_('error occurred')))
            return
        if 'values' in build.result:
            for value in build.result['values']:
                LOG.info(self._translate(
                    _('Value = ${v}', {'v': value})))
        if 'files' in build.result:
            for name in build.result['files']:
                LOG.info(self._translate(_(
                    'File = ${n}', {'n': join(build.path, 'Output', name)})))

    # -------------------------------------------------------------------------
    def _list_processors(self, build_manager):
        """List available processors.

        :param build_manager: (:class:`~.lib.build.agent.AgentBuildManager`
            instance) Current build_manager objet.
        :return: (boolean)
            ``True`` if processors have been listed.
        """
        if not self._opts.list_processors:
            return False

        print '=' * 60
        print '{0:^60}'.format(self._translate(_('Processor list')))
        print '=' * 60

        lang = getdefaultlocale()[0]
        for processor_id in build_manager.processor_list():
            tree = load(join(build_manager.processor_path(processor_id),
                             'processor.xml'), self._relaxngs)
            if isinstance(tree, basestring):
                print u'[{0:<16}] !!! {1:}'.format(
                    processor_id, self._translate(tree))
                continue

            # pylint: disable = I0011, E1103
            print u'[{0:<16}] {1:}'.format(processor_id,
                local_text(tree, 'processor/label', lang=lang))

        return True

    # -------------------------------------------------------------------------
    def _display_variables(self, processing):
        """Display variable values and contraints.

        :param processing: (dictionary)
            Processing dictionary.
        :return: (boolean)
            ``True`` if variables have been displayed.
        """
        if not self._opts.list_variables and not self._opts.show_variables:
            return False

        fmt = u'  | {0:<11} : {1:}'
        lang = getdefaultlocale()[0]
        variables = processing['variables']
        for group in self._processor.find('processor/variables')\
                .iterchildren(tag=etree.Element):
            print '=' * 60
            print '{0:^60}'.format(local_text(group, 'label', lang=lang))
            print '=' * 60
            if self._opts.show_variables \
                   and group.find('description') is not None:
                print self._format_description(lang, group)
                print '-' * 60

            for var in group.findall('var'):
                var_type = var.get('type')
                value = variables.get(var.get('name'), '')
                print u'{0:} = {1:}'.format(var.get('name'), value)
                if not self._opts.show_variables:
                    continue
                print '  v'
                if var.findtext('default') is not None:
                    print fmt.format(self._translate(_('Default')),
                        var.findtext('default').strip())
                print fmt.format(self._translate(_('Type')), var_type)
                if var_type == 'regex':
                    print fmt.format(self._translate(_('Pattern')),
                        var.findtext('pattern').strip())
                elif var_type == 'select':
                    print fmt.format(self._translate(_('Options')),
                        ', '.join(['[%s] %s' % (k.get('value', k.text), k.text)
                                   for k in var.findall('option')]))
                if  var.find('description') is not None:
                    print fmt.format(self._translate(_('Description')),
                        self._format_description(lang, var, 18))
                print

        return True

    # -------------------------------------------------------------------------
    def _read_settings(self, build_file, build_tree):
        """Read settings from ``build_tree``.

        :param build_file: (string)
             Path to build XML file.
        :param build_tree: (etree.ElementTree)
            XML build tree.
        :return: (dictionary)
             A settings dictionary.
        """
        settings = {}
        element = build_tree.find('build/settings')
        for child in element.iterchildren(tag=etree.Element):
            key = child.get('key')
            if '.root' in key:
                settings[key] = abspath(
                    join(dirname(build_file), child.text.strip()))
            else:
                settings[key] = child.text.strip()

        if self._opts.storage_root:
            settings['storage.root'] = \
                expanduser(self._opts.storage_root)
        if self._opts.build_root:
            settings['build.root'] = expanduser(self._opts.build_root)

        if not settings['build.root']:
            LOG.critical(
                self._translate(_('Must have a directory for builds.')))
        if not 'processor.available' in settings:
            settings['processor.available'] = '*'

        return settings

    # -------------------------------------------------------------------------
    def _read_processing(self, settings, build_manager, build_tree):
        """Load processing structure from ``build_tree``.

        :param settings: (dictionary)
             Settings.
        :param build_manager: (:class:`~.lib.build.agent.AgentBuildManager`
            instance) Current build_manager objet.
        :param build_tree: (etree.ElementTree)
            XML build tree.
        :return: (dictionary)
             A processing structure or ``None`` if fails.
        """
        # Main structure
        root_elt = build_tree.find('build/processing')
        processing = {'processor_id': root_elt.findtext('processor').strip()}

        # Load processor
        # pylint: disable = I0011, E1103
        self._processor = build_manager.processor_path(
            processing['processor_id'])
        if self._processor is None:
            LOG.critical(self._translate(_('Unknown processor ${p}.',
                {'p': processing['processor_id']})))
            return
        self._processor = load(
            join(self._processor, 'processor.xml'), self._relaxngs)
        if isinstance(self._processor, basestring):
            LOG.critical(self._translate(self._processor))
            return

        # Variables
        processing['variables'] = self._read_variables(build_tree)
        if processing['variables'] is None:
            return

        # Resource files
        path = settings['storage.root']
        processing['resources'] = self._read_file_set(
            path, build_tree.find('build/processing/resources'))
        if processing['resources'] is None:
            return

        # Template files
        processing['templates'] = self._read_file_set(
            path, build_tree.find('build/processing/templates'))
        if processing['templates'] is None:
            return

        return processing

    # -------------------------------------------------------------------------
    def _read_pack(self, settings, build_tree, files):
        """Load pack structure from ``build_tree``.

        :param settings: (dictionary)
             Settings.
        :param build_tree: (etree.ElementTree)
            XML build tree.
        :param files: (list)
             List of files on command-line to process.
        :return: (dictionary)
             A pack structure or ``None`` if fails.
        """
        # Main structure
        pack = {'recursive': '0'}
        element = build_tree.find('build/pack')
        if element is not None and element.get('recursive') == 'true'\
               or self._opts.recursive:
            pack['recursive'] = '1'

        # Files
        path = settings['storage.root']
        pack['files'] = self._read_file_set(
            path, build_tree.find('build/pack/files'))
        if pack['files'] is None:
            return

        # Files from command-line
        for name in files:
            name = expanduser(name)
            if not exists(name):
                LOG.critical(self._translate(
                    _('Unknown file "${n}".', {'n': name})))
                return
            pack['files'].append(relpath(name, path))

        # Resource files
        pack['resources'] = self._read_file_set(
            path, build_tree.find('build/pack/resources'))
        if pack['resources'] is None:
            return

        # Template files
        pack['templates'] = self._read_file_set(
            path, build_tree.find('build/pack/templates'))
        if pack['templates'] is None:
            return

        return pack

    # -------------------------------------------------------------------------
    def _read_variables(self, build_tree):
        """Read variable definitions in processor tree and fill ``variables``
        dictionary.

        :param build_tree: (:class:`lxml.etree.ElementTree`)
             Build element tree.
        :return: (dictionary)
             Variable dictionary.
        """
        # pylint: disable = I0011, E1103
        variables = {}
        var = build_tree.find('build/processing/variables')
        values = dict([(k.get('name'), k.findtext('value') is not None
                        and k.findtext('value').strip() or '')
                       for k in var.findall('var')])

        for var in self._processor.findall('processor/variables/group/var'):
            name = var.get('name')
            error = self._translate(_('${v}: bad value.', {'v': name}))
            value = values[name] if name in values \
                    else var.findtext('default') is not None \
                    and var.findtext('default').strip() or ''

            if var.get('type') == 'boolean':
                if not value in ('true', 'false', '1', '0', ''):
                    LOG.critical(error)
                    return
                value = bool(value == 'true' or value == '1')
            elif var.get('type') == 'integer':
                if not value.isdigit():
                    LOG.critical(error)
                    return
                value = int(value)
            elif var.get('type') == 'select':
                if not value in [k.get('value') or k.text
                                 for k in var.findall('option')]:
                    LOG.critical(error)
                    return
                value = int(value) if value.isdigit() else value
            elif var.get('type') == 'regex':
                if not re.match(var.findtext('pattern').strip(), value):
                    LOG.critical(error)
                    return
                value = int(value) if value.isdigit() else value
            variables[name] = value

        return variables

    # -------------------------------------------------------------------------
    def _read_file_set(self, storage_root, element):
        """Read files from ``element``.

        :param storage_root: (string)
            Absolute root path to storages.
        :param element: (etree.Element object)
            ``templates`` XML element.
        :return: (list)
            A list of tuples such as ``(<input_file>, <output_path>)`` or
            ``None`` if fails.
        """
        set_list = []
        if element is None:
            return set_list

        for child in element.iterchildren(tag=etree.Element):
            name = child.text.strip()
            if not exists(join(storage_root, name)):
                LOG.critical(self._translate(
                    _('Unknown file "${n}".', {'n': name})))
                return
            if child.get('to'):
                set_list.append((name, child.get('to')))
            else:
                set_list.append(name)

        return set_list

    # -------------------------------------------------------------------------
    @classmethod
    def _format_description(cls, lang, root_elt, indent=0, width=60):
        """Return formatted description text.

        :param lang: (string)
            Preferred language.
        :param root_elt: (:class:`lxml.etree.Element` instance)
            Description element parent.
        :param indent: (integer, default=0)
            Indent value.
        :param width: (integer, default=60)
            Text width.
        :return: string
        """
        text = ' '.join(local_text(root_elt, 'description', lang=lang).split())
        if not indent and ' --' in text:
            return '\n'.join([fill(k.strip(), width, subsequent_indent='    ')
                              for k in text.split(' --')])

        subsequent_indent = '  |' + ' ' * (indent - 3)
        return fill(text, width, initial_indent=' ' * indent,
                    subsequent_indent=subsequent_indent).strip()

    # -------------------------------------------------------------------------
    @classmethod
    def _translate(cls, text):
        """Return ``text`` translated.

        :param text: (string)
            Text to translate.
        """
        return localizer(getdefaultlocale()[0]).translate(text)


# =============================================================================
if __name__ == '__main__':
    main()



# $Id: __init__.py 0f769dc832a4 2012/01/16 18:23:12 patrick $
# pylint: disable = I0011, C0103
"""Here are defined database main objects and constants."""

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker

DBSession = scoped_session(sessionmaker())
Base = declarative_base()

ID_LEN = 32
LABEL_LEN = 64
DESCRIPTION_LEN = 255
PATH_LEN = 255
VALUE_LEN = 128
ALL_LANG_LIST = ('en', 'fr', 'es')


# =============================================================================
def close_dbsession(request):
    """Call back function to close database session.

    :param request: (:class:`pyramid.request.Request` instance)
        Current request.
    """
    # pylint: disable = I0011, W0613
    DBSession.close()



# $Id: test_models_groups.py 7af639d5ec56 2012/03/18 16:06:23 patrick $
# -*- coding: utf-8 -*-
# pylint: disable = I0011, R0904
"""Tests of ``models.groups`` classes."""

import unittest

from ..tests import ModelTestCase


# =============================================================================
class UnitTestModelsGroupsGroup(ModelTestCase):
    """Unit test class for ``models.groups.Group``."""
    # pylint: disable = I0011, C0103

    _group_id = '_test_group'

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models.groups import Group
        self.session.query(Group)\
            .filter_by(group_id=self._group_id).delete()
        self.session.commit()
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make a ``Group`` object."""
        from ..models.groups import Group
        group = Group(self._group_id, u'Les éditeurs', u"Un groupe d'éditeurs")
        return group

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.groups.Group.__init__]"""
        from ..models.groups import Group
        group1 = self._make_one()
        self.session.add(group1)
        self.session.commit()
        group2 = self.session.query(Group)\
                     .filter_by(group_id=group1.group_id).first()
        self.assertNotEqual(group2, None)
        self.assertEqual(group2.group_id, group1.group_id)
        self.assertEqual(group2.label, group1.label)
        self.assertEqual(group2.description, group1.description)

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.groups.Group.__repr__]"""
        group = self._make_one()
        info = repr(group)
        self.assertTrue('<Group(' in info)
        self.assertTrue("'%s'" % self._group_id in info)
        self.assertTrue("'%s'" % group.label.encode('utf8') in info)


# =============================================================================
class UnitTestModelsGroupsGroupPerm(unittest.TestCase):
    """Unit test class for ``models.groups.GroupPerm``."""

    _group_id = '_test_group'

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make a ``GroupPerm`` object."""
        from ..models.groups import GroupPerm
        return GroupPerm(self._group_id, 'usr', 'manager')

    # -------------------------------------------------------------------------
    def test_constructor(self):
        """[u:models.groups.GroupPerm.__init__]"""
        group_perm = self._make_one()
        self.assertEqual(group_perm.scope, 'usr')
        self.assertEqual(group_perm.level, 'manager')

    # -------------------------------------------------------------------------
    def test_repr(self):
        """[u:models.groups.GroupPerm.__repr__]"""
        group_perm = self._make_one()
        info = repr(group_perm)
        self.assertTrue('<GroupPerm(' in info)
        self.assertTrue("'%s'" % self._group_id in info)
        self.assertTrue("'%s'" % group_perm.scope in info)
        self.assertTrue("'%s'" % group_perm.level in info)


# =============================================================================
class IntegrationTestModelsGroups(ModelTestCase):
    """Integration test class for ``models.groups``."""
    # pylint: disable = I0011, C0103

    _user_id = 2000
    _group_id = '_test_group'

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models.users import User
        from ..models.groups import Group
        self.session.query(User).filter_by(user_id=self._user_id).delete()
        self.session.query(Group).filter_by(group_id=self._group_id).delete()
        self.session.commit()
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_group(self):
        """Make a ``Group`` object."""
        # pylint: disable = I0011, E1101
        from ..models.groups import Group
        group = Group(self._group_id, u'Les éditeurs', u"Un groupe d'éditeurs")
        return group

    # -------------------------------------------------------------------------
    def _make_user(self):
        """Make an ``User`` object."""
        from ..models.users import User
        user = User({'auth.key': 'seekrit'}, '_test_user', status='active',
                    password='mypassword', name=u'Félicie TASSION',
                    email='test@prismallia.fr', lang='en')
        user.user_id = self._user_id
        return user

    # -------------------------------------------------------------------------
    def test_delete_cascade_perm(self):
        """[i:models.groups] delete, cascade permission"""
        from ..models.groups import Group, GroupPerm
        group = self._make_group()
        group.perms.append(GroupPerm(self._group_id, 'usr', 'manager'))
        self.session.add(group)
        self.session.commit()
        group_perm = self.session.query(GroupPerm)\
            .filter_by(group_id=self._group_id).first()
        self.assertNotEqual(group_perm, None)
        self.assertEqual(group_perm.scope, 'usr')
        self.assertEqual(group_perm.level, 'manager')
        self.session.query(Group)\
            .filter_by(group_id=self._group_id).delete('fetch')
        self.session.commit()
        group_perm = self.session.query(GroupPerm)\
            .filter_by(group_id=self._group_id).first()
        self.assertEqual(group_perm, None)

    # -------------------------------------------------------------------------
    def test_remove_user(self):
        """[i:models.groups] remove user"""
        from ..models.groups import Group
        user = self._make_user()
        group = self._make_group()
        group.users.append(user)
        self.session.add(user)
        self.session.add(group)
        self.session.commit()
        group = self.session.query(Group)\
                .filter_by(group_id=self._group_id).first()
        self.assertEqual(len(group.users), 1)
        del group.users[0]
        self.session.commit()
        group = self.session.query(Group)\
                .filter_by(group_id=self._group_id).first()
        self.assertEqual(len(group.users), 0)

    # -------------------------------------------------------------------------
    def test_delete_user(self):
        """[i:models.groups] delete user"""
        from ..models.users import User
        user = self._make_user()
        self.session.add(user)
        group = self._make_group()
        group.users.append(user)
        self.session.add(group)
        self.session.commit()
        self.session.query(User)\
            .filter_by(user_id=self._user_id).delete()
        self.session.commit()
        self.assertEqual(len(group.users), 0)



# $Id: test_lib_renderer.py 91a251074d63 2012/03/18 23:07:53 patrick $
# pylint: disable = I0011, R0904
"""Tests of ``lib.renderer`` classes and functions."""

import unittest

from pyramid import testing

from ..tests import ModelTestCase


# =============================================================================
class UnitTestLibRendererHasPermission(unittest.TestCase):
    """Unit test class for ``lib.renderer.has_permission``."""

    # -------------------------------------------------------------------------
    def test_unauthenticated(self):
        """[u:lib.renderer.has_permission] unauthenticated"""
        from ..lib.renderer import has_permission
        request = testing.DummyRequest()
        self.assertFalse(has_permission(request, 'usr_user'))

    # -------------------------------------------------------------------------
    def test_with_admin(self):
        """[u:lib.renderer.has_permission] with administrator"""
        from ..lib.renderer import has_permission
        request = testing.DummyRequest()
        request.session['perms'] = ('admin',)
        self.assertTrue(has_permission(request, 'usr_manager'))

    # -------------------------------------------------------------------------
    def test_with_user(self):
        """[u:lib.renderer.has_permission] with normal user"""
        from ..lib.renderer import has_permission
        request = testing.DummyRequest()
        request.session['perms'] = ('usr_manager', 'grp_editor')
        self.assertTrue(has_permission(request, 'usr_user'))
        self.assertTrue(has_permission(request, 'usr_editor'))
        self.assertTrue(has_permission(request, 'usr_manager'))
        self.assertTrue(has_permission(request, 'grp_user'))
        self.assertTrue(has_permission(request, 'grp_editor'))
        self.assertFalse(has_permission(request, 'grp_manager'))
        self.assertFalse(has_permission(request, 'stg_user'))
        self.assertFalse(has_permission(request, 'stg_editor'))
        self.assertFalse(has_permission(request, 'stg_manager'))


# =============================================================================
class UnitTestLibRendererMenu(ModelTestCase):
    """Unit test class for ``lib.renderer.Menu``."""
    # pylint: disable = I0011, C0103, R0904, W0212

    _key = 'seekrit'
    _user_id = 1000
    _storage_id = '_test_storage'
    _project_id = 1000

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        from ..models.users import User
        from ..models.projects import Project
        from ..models.storages import Storage
        self.session.query(User).filter_by(user_id=self._user_id).delete()
        if self.session.query(Storage)\
               .filter_by(storage_id=self._storage_id).first() is not None:
            Storage.delete('/tmp', self._storage_id)
        self.session.query(Project)\
            .filter_by(project_id=self._project_id).delete()
        self.session.commit()
        ModelTestCase.tearDown(self)

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Make a ``Menu`` object."""
        from ..lib.renderer import Menu
        from ..models.users import User
        from ..models.storages import Storage, StorageUser
        from ..models.projects import Project, ProjectUser
        # User, storage and project
        user = User({'auth.key': self._key}, '_test_user', status='active',
                    password='pass', name=u'Marc', email='test@prismallia.fr',
                    lang='en')
        user.user_id = self._user_id
        self.session.add(user)
        storage = Storage(
            {'auth.key': self._key}, self._storage_id, 'open', 'local')
        storage.users.append(StorageUser(
            self._storage_id, user.user_id, True))
        self.session.add(storage)
        project = Project(u'My project', 'active')
        project.project_id = self._project_id
        project.users.append(ProjectUser(
            project.project_id, self._user_id, True))
        self.session.add(project)
        self.session.commit()
        # Routes
        self.config.add_route('user_admin', '/user/admin')
        self.config.add_route('group_admin', '/group/admin')
        self.config.add_route('storage_admin', '/storage/admin')
        self.config.add_route('storage_index', '/storage/index')
        self.config.add_route('storage_root',
                              '/storage/browse/{storage_id}')
        self.config.add_route('storage_browse',
                              '/storage/browse/{storage_id}/*path')
        self.config.add_route('file_search', '/file/search')
        self.config.add_route('project_admin', '/project/admin')
        self.config.add_route('project_index', '/project/index')
        self.config.add_route('project_view', '/project/view/{project_id}')
        self.config.add_route('project_dashboard',
                              '/project/dashboard/{project_id}')
        self.config.add_route('project_tasks', '/project/tasks/{project_id}')
        self.config.add_route('project_results',
                              '/project/results/{project_id}')
        self.config.add_route('project_doc', '/project/doc/{project_id}')
        # Dummy request
        request = testing.DummyRequest()
        request.session['perms'] = (
            'stg_user', 'prj_user', 'usr_manager', 'grp_editor')
        request.session['lang'] = 'fr'
        request.session['user_id'] = self._user_id
        request.session['breadcrumbs'] = (('Home', 'home', {}, 1),
            ('All storages', 'storage_index', {}, 6))
        request.registry.settings['pyramid.default_locale_name'] = 'en'
        return Menu(request)

    # -------------------------------------------------------------------------
    def test_xhtml(self):
        """[u:lib.renderer.Menu.xhtml]"""
        menu = self._make_one()
        xhtml = menu.xhtml()
        self.assertTrue(xhtml.startswith('<ul>'))
        self.assertTrue('<strong>Storages</strong>' in xhtml)
        self.assertTrue('href="/storage/index"' in xhtml)
        self.assertTrue(
            'href="/storage/browse/%s"' % self._storage_id in xhtml)
        self.assertTrue('<strong>Projects</strong>' in xhtml)
        self.assertTrue('href="/project/index"' in xhtml)
        self.assertTrue(
            'href="/project/dashboard/%s"' % self._project_id in xhtml)
        self.assertTrue('<strong>Administration</strong>' in xhtml)
        self.assertTrue('href="/user/admin"' in xhtml)
        self.assertTrue('href="/group/admin"' in xhtml)
        self.assertFalse('href="/storage/admin"' in xhtml)
        self.assertFalse('href="/project/admin"' in xhtml)
